DUBLINCORE = {'contributor', 'coverage', 'creator',
              'date', 'description', 'format', 'identifier',
              'language', 'publisher', 'relation', 'rights',
              'source', 'subject', 'title', 'type'}
TAGGING = {'keywords', 'tags', 'categories'}
ALL_METADATA = DUBLINCORE.union(TAGGING)

